#!/usr/bin/env python

"""
Author: Ran Wei, Myunghwa Hwang
"""

import pysal
import numpy as np

class Simulation(object):
   
    def __init__(self, src_filename):
        "create a bidirectional network with its total length and total number of links"
        self.nw = pysal.open(src_filename, 'r')
        self.G = {} # {edge_index:(d12,(n1,n2))}
        self.GNet = {} # {n1:{n2:d12}}
        self.total_length = 0.0
        self.nwNum = 0
        for line in self.nw:
            vertices = line.vertices
            for i, vertex in enumerate(vertices[:-1]):
                n1, n2 = vertex, vertices[i+1]
                self.G.setdefault(self.nwNum, ())
                self.GNet.setdefault(n1, {})
                self.GNet.setdefault(n2, {})
                d = pysal.cg.get_points_dist(pysal.cg.Point(n1), pysal.cg.Point(n2))
                self.G[self.nwNum] = (d, (n1, n2))
                self.GNet[n1][n2] = self.nwNum
                self.GNet[n2][n1] = self.nwNum
                self.total_length += d
                self.nwNum += 1
        self.nw.close()
        self.imaginaryLineGenerated = False
            
    def generateImaginaryLine(self):
        '''
           Create an imaginary line that starts from 0 and ends at 1
           and mark the locations of end points of each link
        '''
        self.nwCumPro = [0.0]
        for e in list(self.G.keys()):
            self.nwCumPro.append(self.nwCumPro[-1] + (self.G[e][0]/self.total_length))
                                                                        
        # self.nwCumProDict --> {edge_index:right_side_end_point_of_the_link_on_the_imaginary_line}
        self.nwCumProDict = dict(list(zip(list(self.G.keys()), self.nwCumPro[1:]))) 
        self.imaginaryLineGenerated = True
            
    def getRandomPoints(self, n, projected=False, toShp=False):   
        '''Create a random point pattern data set on the given network'''

        if not self.imaginaryLineGenerated:
            self.generateImaginaryLine()

        ## generate n unique random numbers between 0 and 1
        #randSet = set()
        #while len(randSet) < n: 
        #    randSet = set(np.random.random_sample(n))
        #randSet = np.array(list(randSet))        

        # generate n random numbers between 0 and 1
        randSet = np.random.random_sample(n)

        # Assign the random numbers to the links on the network
        # Think nwCumPro as bins; get bin numbers for all random numbers
        randSet_to_bins=np.digitize(randSet,self.nwCumPro)
        randSet_to_bins=list(zip(randSet_to_bins,randSet))
        randSet_to_bins.sort()
        # Determine geographic coordinates for each random number
        nwPtDict = {}
        for bin_id, rand_number in randSet_to_bins:
            bid = bin_id - 1
            n1, n2 = self.G[bid][1] # n1 and n2 are geographic (real) coordinates for the end points of a link
            origin = 0 if bid <= 0 else self.nwCumProDict[bid-1]
            length = self.nwCumProDict[bid] - origin
            # get prop to determine the geographic coordinate of a random number on the link (n1, n2)
            # length is the length of the link (n1, n2) on the imaginary line
            # (self.nwCumProDict[bin_id] - rand_number) is the distance between a random point and n2 
            # on the imaginary line
            prop = (self.nwCumProDict[bid] - rand_number)*1.0/length
            nwPtDict.setdefault(bid, [])
            if not projected:
                x = n2[0] - (n2[0] - n1[0])*prop # n2[0]: the geographic coordinate of n2 on the X axis
                y = n2[1] - (n2[1] - n1[1])*prop # n2[1]: the geographic coordinate of n2 on the Y axis
                nwPtDict[bid].append((x,y))
            else:
                dist = self.G[bid][0]
                proj_pnt = (n1, n2, dist*(1-prop), dist*prop)
                if toShp:
                    x = n2[0] - (n2[0] - n1[0])*prop
                    y = n2[1] - (n2[1] - n1[1])*prop
                    proj_pnt = tuple(list(proj_pnt) + [x, y])                        
                nwPtDict[bid].append(proj_pnt)
            
        return nwPtDict

    def countPointsOnNetwork(self, points, defaultBase=True):
        G = {}
        for k in self.G:
            n1, n2 = self.G[k][-1]
            G.setdefault(n1, {})
            G.setdefault(n2, {})
            if n2 not in G[n1]:
                attr = G[n1].setdefault(n2, [self.G[n1][n2], 0])
            if n1 not in G[n2]:
                attr = G[n2].setdefault(n1, [self.G[n2][n1], 0])
            if k in points:
                attr[-1] += len(points[k])
            if defaultBase:
                attr += [1.0]
            G[n1][n2] = attr
            G[n2][n1] = attr
        return G

    def createProjRandomPointsShp(self, n, out_filename):
        points = nwPtDict = self.getRandomPoints(n, projected=True, toShp=True)
        shp = pysal.open(out_filename, 'w')
        dbf = pysal.open(out_filename[:-3] + 'dbf', 'w')
        dbf.header = ['ID', 'FROM_P1', 'FROM_P2', 'TO_P1', 'TO_P2', 'D_FROM', 'D_TO']
        dbf.field_spec = [('N',9,0)] + [('N',18,7)]*6
        counter = 0
        for k in points:
            for p in points[k]:
                p = list(p)
                shp.write(pysal.cg.Point(tuple(p[-2:])))
                dbf.write([counter, p[0][0], p[0][1], p[1][0], p[1][1], p[2], p[3]])
                counter += 1
        shp.close()
        dbf.close()
            
    def createRandomPointsShp(self, n, out_filename):
        nwPtDict = self.getRandomPoints(n)
        self.writePoints(nwPtDict, out_filename)

    def writePoints(self, points, out_filename):
        shp = pysal.open(out_filename, 'w')
        dbf = pysal.open(out_filename[:-3] + 'dbf', 'w')
        dbf.header = ['ID']
        dbf.field_spec = [('N',9,0)]
        counter = 0
        for k in points:
            for p in points[k]:
                shp.write(pysal.cg.Point(tuple(p)))
                dbf.write([counter])
                counter += 1
        shp.close()
        dbf.close()

    def getClusteredPoints(self, centerNum, ptNum, percent, clusterMeta=False):

        # split network into center- and non-center network
        centerIDs = np.random.randint(0, self.nwNum, centerNum)
        centers = set(centerIDs)
        centerG, centerG_length = {}, 0
        counter, counter2ID = 0, {}
        for center in centerIDs:
            centerG[counter] = self.G[center]
            centerG_length += self.G[center][0]
            counter2ID[counter] = center
            counter += 1
            n1, n2 = self.G[center][1]
            for neighbor in self.GNet[n1]:
                if neighbor != n2:
                    nghLink = self.GNet[n1][neighbor]
                    centerG[counter] = self.G[nghLink]
                    centerG_length += self.G[nghLink][0]
                    counter2ID[counter] = nghLink
                    counter += 1
                    centers.add(nghLink)
            for neighbor in self.GNet[n2]:
                if neighbor != n1:
                    nghLink = self.GNet[n2][neighbor]
                    centerG[counter] = self.G[nghLink]
                    centerG_length += self.G[nghLink][0]
                    counter2ID[counter] = nghLink
                    counter += 1
                    centers.add(nghLink)
        nonCenterIDs = set(self.G.keys()).difference(centers)
        nonCenterG, nonCenterG_length = {}, 0
        for i, nonCenter in enumerate(nonCenterIDs):
             nonCenterG[i] = self.G[nonCenter]
             nonCenterG_length += self.G[nonCenter][0]

        self.oldG, self.old_total_length = self.G, self.total_length

        self.G, self.total_length = centerG, centerG_length
        n_centerPoints = int(percent*ptNum*1.0)
        self.imaginaryLineGenerated = False
        pointsInCenter = self.getRandomPoints(n_centerPoints) 
        meta = {}
        if clusterMeta:
            for cluster in pointsInCenter:
                num_points = len(pointsInCenter[cluster])
                centerLink = self.oldG[cluster][1]
                centerID = counter2ID[cluster]
                centerLink = self.oldG[centerID][1]
                meta[centerLink] = [self.oldG[centerID][0], num_points, centerID in centerIDs]
                #meta[centerID] = [num_points] + list(self.oldG[centerID])
       
        self.G, self.total_length = nonCenterG, nonCenterG_length 
        self.imaginaryLineGenerated = False
        pointsInNonCenter = self.getRandomPoints(ptNum - n_centerPoints)
        centers_no = len(centers)
        for k in pointsInNonCenter:
            pointsInCenter[k + centers_no] = pointsInNonCenter[k]
        pointsInCenter.update(pointsInNonCenter)

        self.G, self.total_length = self.oldG, self.old_total_length

        self.imaginaryLineGenerated = False

        return pointsInCenter, meta

    def writeMeta(self, metaData, out_file):
        shp = pysal.open(out_file, 'w')
        dbf = pysal.open(out_file[:-3] + 'dbf', 'w')
        dbf.header = ['LENGTH', 'NO_PNTS', 'INITIAL_CENTER']
        dbf.field_spec = [('N',9,0)]*2 + [('L',1,0)]
        for link in metaData:
            vertices = list(link)
            vertices = [pysal.cg.Point(v) for v in vertices]
            shp.write(pysal.cg.Chain(vertices))
            dbf.write(metaData[link])
        shp.close()
        dbf.close()

    def createClusteredPointsShp(self, centerNum, n, percent, out_filename, clusterMetaFile=None):
        nwPtDict, meta = self.getClusteredPoints(centerNum, n, percent, clusterMetaFile!=None)
        self.writePoints(nwPtDict, out_filename)
        if clusterMetaFile:
            self.writeMeta(meta, clusterMetaFile)

if __name__ == '__main__':
    sim=Simulation("streets.shp")
    sim.createProjRandomPointsShp(100, "random_100.shp")
    sim.createClusteredPointsShp(2, 100, 0.1, "clustered_100_10p.shp", "clustered_100_10p_meta.shp") 



"""
:mod:`esda` --- Exploratory Spatial Data Analysis
=================================================

"""
from . import mapclassify
from . import moran
from . import smoothing
from . import getisord
from . import geary
from . import join_counts
from . import gamma



"""miscellaneous file manipulation utilities

"""

import numpy as np
import pysal as ps
import pandas as pd

def check_dups(li):
    """checks duplicates in list of ID values
       ID values must be read in as a list

       __author__  = "Luc Anselin <luc.anselin@asu.edu> "
       
       Arguments
       ---------
       li      : list of ID values
       
       Returns
       -------
       a list with the duplicate IDs
    """
    return list(set([x for x in li if li.count(x) > 1]))
    
    
def dbfdups(dbfpath,idvar):
    """checks duplicates in a dBase file
       ID variable must be specified correctly

       __author__  = "Luc Anselin <luc.anselin@asu.edu> " 
       
       Arguments
       ---------
       dbfpath  : file path to dBase file
       idvar    : ID variable in dBase file
       
       Returns
       -------
       a list with the duplicate IDs
    """
    db = ps.open(dbfpath,'r')
    li = db.by_col(idvar)
    return list(set([x for x in li if li.count(x) > 1]))   
    

def df2dbf(df, dbf_path, my_specs=None):
    '''
    Convert a pandas.DataFrame into a dbf.

    __author__  = "Dani Arribas-Bel <darribas@asu.edu>, Luc Anselin <luc.anselin@asu.edu>"
    ...

    Arguments
    ---------
    df          : DataFrame
                  Pandas dataframe object to be entirely written out to a dbf
    dbf_path    : str
                  Path to the output dbf. It is also returned by the function
    my_specs    : list
                  List with the field_specs to use for each column.
                  Defaults to None and applies the following scheme:
                    * int: ('N', 14, 0) - for all ints
                    * float: ('N', 14, 14) - for all floats
                    * str: ('C', 14, 0) - for string, object and category
                  with all variants for different type sizes
                  
    Note: use of dtypes.name may not be fully robust, but preferred apprach of using
    isinstance seems too clumsy
    '''
    if my_specs:
        specs = my_specs
    else:
        """
        type2spec = {int: ('N', 20, 0),
                     np.int64: ('N', 20, 0),
                     np.int32: ('N', 20, 0),
                     np.int16: ('N', 20, 0),
                     np.int8: ('N', 20, 0),
                     float: ('N', 36, 15),
                     np.float64: ('N', 36, 15),
                     np.float32: ('N', 36, 15),
                     str: ('C', 14, 0)
                     }
        types = [type(df[i].iloc[0]) for i in df.columns]
        """
        # new approach using dtypes.name to avoid numpy name issue in type
        type2spec = {'int': ('N', 20, 0),
                     'int8': ('N', 20, 0),
                     'int16': ('N', 20, 0),
                     'int32': ('N', 20, 0),
                     'int64': ('N', 20, 0),
                     'float': ('N', 36, 15),
                     'float32': ('N', 36, 15),
                     'float64': ('N', 36, 15),
                     'str': ('C', 14, 0),
                     'object': ('C', 14, 0),
                     'category': ('C', 14, 0)
                     }
        types = [df[i].dtypes.name for i in df.columns]
        specs = [type2spec[t] for t in types]
    db = ps.open(dbf_path, 'w')
    db.header = list(df.columns)
    db.field_spec = specs
    for i, row in df.T.items():
        db.write(row)
    db.close()
    return dbf_path

def dbf2df(dbf_path, index=None, cols=False, incl_index=False):
    '''
    Read a dbf file as a pandas.DataFrame, optionally selecting the index
    variable and which columns are to be loaded.

    __author__  = "Dani Arribas-Bel <darribas@asu.edu> "
    ...

    Arguments
    ---------
    dbf_path    : str
                  Path to the DBF file to be read
    index       : str
                  Name of the column to be used as the index of the DataFrame
    cols        : list
                  List with the names of the columns to be read into the
                  DataFrame. Defaults to False, which reads the whole dbf
    incl_index  : Boolean
                  If True index is included in the DataFrame as a
                  column too. Defaults to False

    Returns
    -------
    df          : DataFrame
                  pandas.DataFrame object created
    '''
    db = ps.open(dbf_path)
    if cols:
        if incl_index:
            cols.append(index)
        vars_to_read = cols
    else:
        vars_to_read = db.header
    data = dict([(var, db.by_col(var)) for var in vars_to_read])
    if index:
        index = db.by_col(index)
        db.close()
        return pd.DataFrame(data, index=index, columns=vars_to_read)
    else:
        db.close()
        return pd.DataFrame(data,columns=vars_to_read)

    
def dbfjoin(dbf1_path,dbf2_path,out_path,joinkey1,joinkey2):
    '''
    Wrapper function to merge two dbf files into a new dbf file. 

    __author__  = "Luc Anselin <luc.anselin@asu.edu> "
    
    Uses dbf2df and df2dbf to read and write the dbf files into a pandas
    DataFrame. Uses all default settings for dbf2df and df2dbf (see docs
    for specifics).
    ...

    Arguments
    ---------

    dbf1_path   : str
                  Path to the first (left) dbf file
    dbf2_path   : str
                  Path to the second (right) dbf file
    out_path    : str
                  Path to the output dbf file (returned by the function)
    joinkey1    : str
                  Variable name for the key in the first dbf. Must be specified.
                  Key must take unique values.
    joinkey2    : str
                  Variable name for the key in the second dbf. Must be specified.
                  Key must take unique values.
                  
    Returns
    -------
    dbfpath     : path to output file
    
    '''
    df1 = dbf2df(dbf1_path,index=joinkey1)
    df2 = dbf2df(dbf2_path,index=joinkey2)
    dfbig = pd.merge(df1,df2,left_on=joinkey1,right_on=joinkey2,sort=False)
    dp = df2dbf(dfbig,out_path)
    return dp


def dta2dbf(dta_path,dbf_path):
    """
    Wrapper function to convert a stata dta file into a dbf file. 

    __author__  = "Luc Anselin <luc.anselin@asu.edu> "
    
    Uses df2dbf to write the dbf files from a pandas
    DataFrame. Uses all default settings for df2dbf (see docs
    for specifics).
    ...

    Arguments
    ---------

    dta_path    : str
                  Path to the Stata dta file
    dbf_path    : str
                  Path to the output dbf file
                  
    Returns
    -------
    dbf_path    : path to output file
    """
    
    db = pd.read_stata(dta_path)
    dp = df2dbf(db,dbf_path)
    return dp    



import pysal
from pysal.cg.shapes import Point, Chain
import doctest
import unittest


class test_MultiPloygon(unittest.TestCase):

    def test___init__1(self):
        """
        Tests conversion of polygons with multiple shells to 
        geoJSON multipolygons. and back.
        """
        shp = pysal.open(pysal.examples.get_path("NAT.shp"),'r')
        multipolygons = [p for p in shp if len(p.parts) > 1]
        geoJSON = [p.__geo_interface__ for p in multipolygons]
        for poly in multipolygons:
            json = poly.__geo_interface__
            shape = pysal.cg.asShape(json)
            self.assertEqual(json['type'],'MultiPolygon')
            self.assertEqual(str(shape.holes), str(poly.holes))
            self.assertEqual(str(shape.parts), str(poly.parts))
class test_MultiLineString(unittest.TestCase):
    def test_multipart_chain(self):
        vertices = [[Point((0, 0)), Point((1, 0)), Point((1, 5))],
                    [Point((-5, -5)), Point((-5, 0)), Point((0, 0))]]

        #part A
        chain0 = Chain(vertices[0])
        #part B
        chain1 = Chain(vertices[1])
        #part A and B
        chain2 = Chain(vertices)

        json = chain0.__geo_interface__
        self.assertEqual(json['type'], 'LineString')
        self.assertEqual(len(json['coordinates']), 3)

        json = chain1.__geo_interface__
        self.assertEqual(json['type'], 'LineString')
        self.assertEqual(len(json['coordinates']), 3)

        json = chain2.__geo_interface__
        self.assertEqual(json['type'], 'MultiLineString')
        self.assertEqual(len(json['coordinates']), 2)

        chain3 = pysal.cg.asShape(json)
        self.assertEqual(chain2.parts, chain3.parts)


if __name__ == '__main__':
    unittest.main()
    #runner = unittest.TextTestRunner()
    #runner.run(suite)



"""
Geary's C statistic for spatial autocorrelation
"""
__author__ = "Sergio J. Rey <srey@asu.edu> "

import numpy as np
import scipy.stats as stats
from .. import weights
from .tabular import _univariate_handler

__all__ = ['Geary']


class Geary(object):
    """
    Global Geary C Autocorrelation statistic

    Parameters
    ----------
    y              : array
                     (n, 1) attribute vector
    w              : W
                     spatial weights
    transformation : {'B', 'R', 'D', 'U', 'V'}
                     weights transformation, default is binary.
                     Other options include "R": row-standardized, "D":
                     doubly-standardized, "U": untransformed (general
                     weights), "V": variance-stabilizing.
    permutations   : int
                     number of random permutations for calculation of
                     pseudo-p_values

    Attributes
    ----------
    y              : array
                     original variable
    w              : W
                     spatial weights
    permutations   : int
                     number of permutations
    C              : float
                     value of statistic
    EC             : float
                     expected value
    VC             : float
                     variance of G under normality assumption
    z_norm         : float
                     z-statistic for C under normality assumption
    z_rand         : float
                     z-statistic for C under randomization assumption
    p_norm         : float
                     p-value under normality assumption (one-tailed)
    p_rand         : float
                     p-value under randomization assumption (one-tailed)
    sim            : array
                     (if permutations!=0)
                     vector of I values for permutated samples
    p_sim          : float
                     (if permutations!=0)
                     p-value based on permutations (one-tailed)
                     null: sptial randomness
                     alternative: the observed C is extreme
                     it is either extremely high or extremely low
    EC_sim         : float
                     (if permutations!=0)
                     average value of C from permutations
    VC_sim         : float
                     (if permutations!=0)
                     variance of C from permutations
    seC_sim        : float
                     (if permutations!=0)
                     standard deviation of C under permutations.
    z_sim          : float
                     (if permutations!=0)
                     standardized C based on permutations
    p_z_sim        : float
                     (if permutations!=0)
                     p-value based on standard normal approximation from
                     permutations (one-tailed)

    Examples
    --------
    >>> import pysal
    >>> w = pysal.open(pysal.examples.get_path("book.gal")).read()
    >>> f = pysal.open(pysal.examples.get_path("book.txt"))
    >>> y = np.array(f.by_col['y'])
    >>> c = Geary(y,w,permutations=0)
    >>> print round(c.C,7)
    0.3330108
    >>> print round(c.p_norm,7)
    9.2e-05
    >>> c.z_rand
    -3.6489513837253944
    """
    def __init__(self, y, w, transformation="r", permutations=999):
        if not isinstance(w, weights.W):
            raise TypeError('w must be a pysal weights object, got {}'
                            ' instead'.format(type(w)))
        y = np.asarray(y).flatten()
        self.n = len(y)
        self.y = y
        w.transform = transformation
        self.w = w
        self.permutations = permutations
        self.__moments()
        xn = range(len(y))
        self.xn = xn
        self.y2 = y * y
        yd = y - y.mean()
        yss = sum(yd * yd)
        self.den = yss * self.w.s0 * 2.0
        self.C = self.__calc(y)
        de = self.C - 1.0
        self.EC = 1.0
        self.z_norm = de / self.seC_norm
        self.z_rand = de / self.seC_rand
        if de > 0:
            self.p_norm = 1 - stats.norm.cdf(self.z_norm)
            self.p_rand = 1 - stats.norm.cdf(self.z_rand)
        else:
            self.p_norm = stats.norm.cdf(self.z_norm)
            self.p_rand = stats.norm.cdf(self.z_rand)


        if permutations:
            sim = [self.__calc(np.random.permutation(self.y))
                   for i in range(permutations)]
            self.sim = sim = np.array(sim)
            above = sim >= self.C
            larger = sum(above)
            if (permutations - larger) < larger:
                larger = permutations - larger
            self.p_sim = (larger + 1.) / (permutations + 1.)
            self.EC_sim = sum(sim) / permutations
            self.seC_sim = np.array(sim).std()
            self.VC_sim = self.seC_sim ** 2
            self.z_sim = (self.C - self.EC_sim) / self.seC_sim
            self.p_z_sim = 1 - stats.norm.cdf(np.abs(self.z_sim))

    @property
    def _statistic(self):
        """ a standardized accessor for esda statistics"""
        return self.C

    def __moments(self):
        y = self.y
        n = self.n
        w = self.w
        s0 = w.s0
        s1 = w.s1
        s2 = w.s2
        s02 = s0 * s0
        yd = y - y.mean()
        yd4 = yd**4
        yd2 = yd**2
        n2 = n * n
        k = (yd4.sum() / n) / ((yd2.sum()/n)**2)
        A = (n-1) * s1 * (n2 - 3*n + 3 - (n-1) * k)
        B = (1./4) * ((n-1) * s2 * (n2 + 3*n - 6 - (n2 - n +2) * k ))
        C = s02 * (n2 - 3 - (n-1)**2 * k)
        vc_rand = (A-B+C) / (n * (n-2) * (n-3)*s02)
        vc_norm = ((1 / (2 * (n + 1) * s02)) *
                   ((2 * s1 + s2) * (n - 1) - 4 * s02))

        self.VC_rand = vc_rand
        self.VC_norm = vc_norm
        self.seC_rand = vc_rand ** (0.5)
        self.seC_norm = vc_norm ** (0.5)

    def __calc(self, y):
        ys = np.zeros(y.shape)
        y2 = y ** 2
        for i, i0 in enumerate(self.w.id_order):
            neighbors = self.w.neighbor_offsets[i0]
            wijs = self.w.weights[i0]
            z = list(zip(neighbors, wijs))
            ys[i] = sum([wij * (y2[i] - 2 * y[i] * y[j] + y2[j])
                         for j, wij in z])
        a = (self.n - 1) * sum(ys)
        return a / self.den

    @classmethod
    def by_col(cls, df, cols, w=None, inplace=False, pvalue='sim', outvals=None, **stat_kws):
        """ 
        Function to compute a Geary statistic on a dataframe

        Arguments
        ---------
        df          :   pandas.DataFrame
                        a pandas dataframe with a geometry column
        cols        :   string or list of string
                        name or list of names of columns to use to compute the statistic
        w           :   pysal weights object
                        a weights object aligned with the dataframe. If not provided, this
                        is searched for in the dataframe's metadata
        inplace     :   bool
                        a boolean denoting whether to operate on the dataframe inplace or to
                        return a series contaning the results of the computation. If
                        operating inplace, with default configurations, 
                        the derived columns will be named like 'column_geary' and 'column_p_sim'
        pvalue      :   string
                        a string denoting which pvalue should be returned. Refer to the
                        the Geary statistic's documentation for available p-values
        outvals     :   list of strings
                        list of arbitrary attributes to return as columns from the 
                        Geary statistic
        **stat_kws  :   keyword arguments
                        options to pass to the underlying statistic. For this, see the
                        documentation for the Geary statistic.

        Returns
        --------
        If inplace, None, and operation is conducted on dataframe in memory. Otherwise,
        returns a copy of the dataframe with the relevant columns attached.

        See Also
        ---------
        For further documentation, refer to the Geary class in pysal.esda
        """
        return _univariate_handler(df, cols, w=w, inplace=inplace, pvalue=pvalue,
                                   outvals=outvals, stat=cls,
                                   swapname=cls.__name__.lower(), **stat_kws)
